import copy
from collections import namedtuple
from typing import List, Union

import importlib_metadata
import marshmallow as ma
from marshmallow import fields

from ..utils.facet_helpers import facet_definition, facet_name

Import = namedtuple("Import", "import_path,alias")


class DataType:
    model_type = None
    marshmallow_field = None
    ui_marshmallow_field = None
    schema_type = None
    mapping_type = None
    default_facet_class = "TermsFacet"
    default_facet_imports = [
        {"import": "invenio_records_resources.services.records.facets.TermsFacet"}
    ]

    class ModelSchema(ma.Schema):
        type = fields.String(required=True)

    def __init__(self, definition, key, model, schema, stack):
        self.definition = definition
        self.key = key
        self.model = model
        self.schema = schema
        self.stack = stack

    def _copy_definition(self, **extras):
        ret = copy.deepcopy(self.definition)
        for k, v in extras.items():
            if v is not None:
                ret[k] = v
        return ret

    def prepare(self, context):
        """Called at the beginning in model-preprocessing phase,
        should prepare self.definition (add defaults etc).
        Might use the provided context to store cross-node information.

        This call should always be deterministic.
        """
        definition = self.definition.setdefault("marshmallow", {})
        if self.marshmallow_field:
            definition.setdefault("field-class", self.marshmallow_field)
        definition.setdefault("validators", []).extend(self.marshmallow_validators())

        ui = self.definition.setdefault("ui", {})
        definition = ui.setdefault("marshmallow", {})
        if self.ui_marshmallow_field:
            definition.setdefault("field-class", self.ui_marshmallow_field)
        elif self.marshmallow_field:
            definition.setdefault("field-class", self.marshmallow_field)

    def model_schema(self, **_extras):
        return None

    def json_schema(self, **extras):
        return self._copy_definition(type=self.schema_type, **extras)

    def mapping(self, **extras):
        return self._copy_definition(type=self.mapping_type, **extras)

    def marshmallow(self, **extras):
        ret = self.definition.get("marshmallow", {})
        for k, v in extras.items():
            if v is not None:
                ret[k] = v
        return ret

    def ui_marshmallow(self, **extras):
        ret = self.definition["ui"]["marshmallow"]
        for k, v in extras.items():
            if v is not None:
                ret[k] = v
        return ret

    def marshmallow_validators(self) -> List[str]:
        return []

    def imports(self, *extra) -> List[Import]:
        return extra

    def dumper_class(self, data):  # NOSONAR
        return None

    @property
    def facet_class(self):
        facets = self.definition.get("facets", {})
        return facets.get("facet-class", self.default_facet_class)

    @property
    def facet_imports(self):
        facets = self.definition.get("facets", {})
        return facets.get("imports", self.default_facet_imports)

    def get_facet(self, stack, parent_path, path_suffix=None):
        """
        path_suffix - intended to be used from subclasses, such as fulltext+keyword or vocabulary
        """
        key, field, args, path = facet_definition(self)
        local_path = parent_path
        if len(parent_path) > 0 and self.key:
            local_path = parent_path + "." + self.key
        elif self.key:
            local_path = self.key
        if path:
            path = local_path + "." + path
        else:
            path = local_path

        f_name = facet_name(f"{local_path}{path_suffix or ''}")

        if field:
            return [{"facet": field, "path": f_name}]
        else:
            label = f"{local_path}{path_suffix or ''}".replace(".", "/") + ".label"
            if args:
                serialized_args = ", " + ", ".join(args)
            else:
                serialized_args = ""

        return self._get_facet_definition(
            stack,
            self.facet_class,
            f_name,
            path,
            path_suffix or "",
            label,
            serialized_args,
        )

    def _get_facet_definition(
        self, stack, facet_class, facet_name, path, path_suffix, label, serialized_args
    ):
        return [
            {
                "facet": f'{facet_class}(field="{path}{path_suffix}", label=_("{label}"){serialized_args})',
                "path": facet_name,
            }
        ]


class DataTypes:
    def __init__(self) -> None:
        self.datatype_map = {}

    def _prepare_datatypes(self):
        if not self.datatype_map:
            for entry in importlib_metadata.entry_points(
                group="oarepo_model_builder.datatypes"
            ):
                for dt in entry.load():
                    self.datatype_map[dt.model_type] = dt

    def get_datatype(self, data, key, model, schema, stack) -> Union[DataType, None]:
        datatype_class = self.get_datatype_class(data.get("type", None))
        if datatype_class:
            return datatype_class(data, key, model, schema, stack)
        return None

    def get_datatype_class(self, datatype_type):
        self._prepare_datatypes()
        return self.datatype_map.get(datatype_type)

    def facet(self, stack):
        return stack[0].get_facet(stack[1:], "")

    def clear_cache(self):
        self.datatype_map = {}


datatypes = DataTypes()



from __future__ import annotations

import functools
import inspect
import sys
from typing import TYPE_CHECKING, List

from oarepo_model_builder.property_preprocessors import PropertyPreprocessor
from oarepo_model_builder.stack import ModelBuilderStack, ReplaceElement
from oarepo_model_builder.utils.json_pathlib import JSONPaths
from oarepo_model_builder.utils.verbose import log

if TYPE_CHECKING:
    from oarepo_model_builder.builder import ModelBuilder


def process(path, priority=0, condition=None):
    def wrapper(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        wrapped.model_builder_path = path
        wrapped.model_builder_priority = priority
        wrapped.model_builder_condition = condition
        return wrapped

    return wrapper


class OutputBuilder:
    TYPE = None
    stack: ModelBuilderStack

    def __init__(
        self, builder: ModelBuilder, property_preprocessors: List[PropertyPreprocessor]
    ):
        self.builder = builder
        self.property_preprocessors = property_preprocessors
        self.stack = None
        self.silent_exceptions = False
        # TODO: move this to metaclass and initialize only once per class
        self.json_paths = JSONPaths()
        arr = []
        for name, method in inspect.getmembers(self, inspect.ismethod):
            if not hasattr(method, "model_builder_priority"):
                continue
            arr.append(
                (
                    -method.model_builder_priority,
                    -len(method.model_builder_path),
                    method.model_builder_path,
                    id(method),
                    method.model_builder_condition,
                    method,
                )
            )
        arr.sort()
        for _prior, _lpath, path, _mid, condition, method in arr:
            self.json_paths.register(path, condition, method)

    def begin(self, schema, settings):
        self.schema = schema.schema
        self.whole_schema = schema
        self.current_model = schema.current_model
        self.settings = settings
        self.stack = ModelBuilderStack()
        self.stack.push(None, schema.current_model)
        log.enter(2, "Creating %s", self.TYPE)
        self.silent_exceptions = False

    def finish(self):
        log.leave()

    def build(self, schema):
        self.begin(schema, schema.settings)

        for proc in self.property_preprocessors:
            proc.begin(schema, schema.settings)

        self.build_children()

        for proc in self.property_preprocessors:
            proc.finish()

        self.finish()

    def build_node(self, key, data):
        try:
            self.stack.push(key, data)

            try:
                for property_preprocessor in self.property_preprocessors:
                    data = (
                        property_preprocessor.process(self.TYPE, data, self.stack)
                        or data
                    )
            except ReplaceElement as e:
                data = e
            if isinstance(data, ReplaceElement):
                self.stack.pop()
                if data.data is not None:
                    if isinstance(data.data, dict):
                        for k, v in data.data.items():
                            self.build_node(k, v)
                    elif isinstance(data.data, (list, tuple)):
                        for k, v in enumerate(data.data):
                            self.build_node(k, v)
                    else:
                        raise AttributeError(
                            f"Do not know how to handle {type(data.data)} in ReplaceElement"
                        )
                return
            self.stack.top.data = data
            self.process_stack_top()
            self.stack.pop()
        except Exception as e:
            if not self.silent_exceptions:
                self.silent_exceptions = True
                print(f"Error on handling path {self.stack.path}: {e}", file=sys.stderr)
            raise

    def build_children(self):
        data = self.stack.top.data
        if isinstance(data, (list, tuple)):
            for k, v in enumerate(data):
                self.build_node(k, v)
        elif isinstance(data, dict):
            children = list(data.items())
            for k, v in children:
                self.build_node(k, v)

    def process_stack_top(self):
        try:
            self.call_components(
                "before_process_element", value=self.stack.top.data, stack=self.stack
            )
            for method in self.json_paths.match(
                self.stack.path, self.stack.top.data, extra_data={"stack": self.stack}
            ):
                return method()
            # do not skip stack top
            if self.stack.level <= 1:
                self.build_children()
        finally:
            self.call_components(
                "after_process_element", value=self.stack.top.data, stack=self.stack
            )

    @process("/model")
    def enter_model(self):
        self.build_children()

    def call_components(self, method_name, value, **kwargs):
        for component in self.builder.get_output_builder_components(self.TYPE):
            if hasattr(component, method_name):
                value = getattr(component, method_name)(self, value, **kwargs) or value
        return value


class OutputBuilderComponent:
    def before_process_element(
        self, builder: OutputBuilder, value, *, stack: ModelBuilderStack, **kwargs
    ):
        return value

    def after_process_element(
        self, builder: OutputBuilder, value, *, stack: ModelBuilderStack, **kwargs
    ):
        return value


TEMPLATES = {
    "setup_py": "templates/setup.py.jinja2",
}

__all__ = [
    "process",
    "OutputBuilder",
    "ModelBuilderStack",
    "ReplaceElement",
    "TEMPLATES",
]



import re

import libcst as cst

from oarepo_model_builder.utils.cst import PythonContext, merge


def test_new_class():
    existing_module = "# comment start"
    included_module = """
# comment before
class Blah:
    # comment
    pass    
    """.strip()
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert (
        transformed_cst.code.strip()
        == """
# comment start
class Blah:
    # comment
    pass    
    """.strip()
    )


def test_existing_class():
    existing_module = """
# comment start
class Blah:
    # comment
    pass    
    """.strip()
    included_module = existing_module

    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert transformed_cst.code.strip() == existing_module.strip()


def test_new_function():
    existing_module = "# comment start"
    included_module = """
# comment before
def a():
    # comment
    return 1
        """.strip()
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert (
        transformed_cst.code.strip()
        == """
# comment start
def a():
    # comment
    return 1    
    """.strip()
    )


def test_existing_function():
    existing_module = """
# comment start
def a():
    # comment
    return 1
        """.strip()
    included_module = """
def a():
    return 2
        """.strip()

    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert transformed_cst.code.strip() == existing_module.strip()


def test_new_method():
    existing_module = """
# comment start
class Blah:
    # comment before
    def b(self):
        return 2
        """.strip()
    included_module = """
class Blah:
    # comment 2 before
    def a(self):
        return 1
        """.strip()

    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert (
        transformed_cst.code.strip()
        == """
# comment start
class Blah:
    # comment before
    def b(self):
        return 2
    # comment 2 before
    def a(self):
        return 1 
    """.strip()
    )


def test_existing_method():
    existing_module = """
# comment start
class Blah:
    # comment before
    def b(self):
        return 2
        """.strip()
    included_module = """
class Blah:
    # comment 2 before
    def b(self):
        return 1
        """.strip()

    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert (
        transformed_cst.code.strip()
        == """
# comment start
class Blah:
    # comment before
    def b(self):
        return 2
    """.strip()
    )


def test_new_include():
    existing_module = "# comment start"
    included_module = """
# comment before
import pathlib
from pathlib import (
    Blah
)
# comment after
class Blah:
    pass
        """.strip()
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert (
        transformed_cst.code.strip()
        == """
# comment start
import pathlib
from pathlib import (
    Blah
)
# comment after
class Blah:
    pass
        """.strip()
    )


def test_existing_include():
    existing_module = """
# comment start
from c import d
import b
""".strip()
    included_module = """
# comment before
import b
import pathlib
from pathlib import (
    Blah
)
from c import d
from c import d, e
            """.strip()
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert (
        transformed_cst.code.strip()
        == """
# comment start
from c import d
import b
import pathlib
from pathlib import (
    Blah
)
from c import d, e
            """.strip()
    )


def test_top_level_vars():
    existing_module = ""
    included_module = """
AAA = "123"
BBB = "456"
            """.strip()
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert (
        transformed_cst.code.strip()
        == """
AAA = "123"
BBB = "456"
            """.strip()
    )


def test_top_level_merged_vars():
    existing_module = "CCC='456'"
    included_module = """
AAA = "123"
BBB = "456"
            """.strip()
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert (
        transformed_cst.code.strip()
        == """
CCC='456'
AAA = "123"
BBB = "456"
            """.strip()
    )


def test_merge_top_level_arrays():
    existing_module = "CCC=[1,2,3]"
    included_module = """CCC=[4,5,6,1]"""
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert (
        transformed_cst.code.strip().replace(" ", "")
        == """
CCC=[1,2,3,4,5,6,]
            """.strip()
    )


def test_do_not_overwrite_top_level_vars():
    existing_module = "CCC=234"
    included_module = """CCC=438"""
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert (
        transformed_cst.code.strip()
        == """
CCC=234
            """.strip()
    )


def test_merge_class_arrays():
    existing_module = """
class A:
    CCC=[1,2,3]
    """
    included_module = """
class A:
    CCC=[4,5,6]
    """
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert re.sub(r"[\t\n ]", "", transformed_cst.code) == re.sub(
        r"[\t\n ]",
        "",
        """
class A:
CCC=[1,2,3,4,5,6]
            """,
    )


def test_merge_arguments():
    existing_module = """
a = blah(1, 2, 3)    
    """
    included_module = """
a = blah(1, 2, 4)   
"""
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert re.sub(r"[\t\n ]", "", transformed_cst.code) == re.sub(
        r"[\t\n ]", "", existing_module
    )


def test_merge_kwarguments():
    existing_module = """
a = blah(1, 2, b=3)    
    """
    included_module = """
a = blah(1, 2, d=4)   
"""
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert re.sub(r"[\t\n ]", "", transformed_cst.code) == re.sub(
        r"[\t\n ]",
        "",
        """
a = blah(1, 2, b=3, d=4)    
    """,
    )


def test_merge_argument_list():
    existing_module = """
a = blah(1, 2, [3])    
    """
    included_module = """
a = blah(1, 2, [4])   
"""
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert re.sub(r"[\t\n ]", "", transformed_cst.code) == re.sub(
        r"[\t\n ]",
        "",
        """
a = blah(1, 2, [3, 4])    
    """,
    )


def test_merge_kwargument_list():
    existing_module = """
a = blah(1, 2, b=[3])    
    """
    included_module = """
a = blah(1, 2, b=[4])   
"""
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert re.sub(r"[\t\n ]", "", transformed_cst.code) == re.sub(
        r"[\t\n ]",
        "",
        """
a = blah(1, 2, b=[3, 4])    
    """,
    )


def test_dict_merge():
    existing_module = """
a = {'a': 1, 'b': 2}  
        """
    included_module = """
a = {'c': 3}
    """
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert re.sub(r"[\t\n ]", "", transformed_cst.code) == re.sub(
        r"[\t\n ]",
        "",
        """
    a = {'a': 1, 'b': 2, 'c': 3}
        """,
    )


def test_dict_merge_overwrite():
    existing_module = """
a = {'a': 1}  
            """
    included_module = """
a = {'a': 2}
        """
    original_cst = cst.parse_module(existing_module)
    included_cst = cst.parse_module(
        included_module, config=original_cst.config_for_parsing
    )
    transformed_cst = merge(PythonContext(included_cst), original_cst, included_cst)

    assert re.sub(r"[\t\n ]", "", transformed_cst.code) == re.sub(
        r"[\t\n ]",
        "",
        """
        a = {'a': 1}
            """,
    )



from oarepo_model_builder.utils.hyphen_munch import HyphenMunch

from .invenio_record_schema import InvenioRecordSchemaBuilder


class InvenioRecordUISchemaBuilder(InvenioRecordSchemaBuilder):
    TYPE = "invenio_record_ui_schema"
    class_config = "record-ui-schema-class"
    template = "record-schema"
    OAREPO_MARSHMALLOW_PROPERTY = "ui_marshmallow"
    extra_imports = [
        HyphenMunch({"import-path": "oarepo_runtime.ui.marshmallow.InvenioUISchema"})
    ]



import copy
import dataclasses
import logging
from collections import defaultdict
from typing import Any, Dict, List, Union

from oarepo_model_builder.builders import process
from oarepo_model_builder.builders.python import PythonBuilder
from oarepo_model_builder.datatypes import Import, ObjectDataType, datatypes
from oarepo_model_builder.schema import ModelSchema
from oarepo_model_builder.stack.stack import ModelBuilderStack
from oarepo_model_builder.utils.jinja import (
    split_base_name,
    split_package_base_name,
    split_package_name,
)
from oarepo_model_builder.utils.python_name import convert_name_to_python
from oarepo_model_builder.validation import InvalidModelException

from .invenio_base import InvenioBaseClassPythonBuilder

log = logging.getLogger("invenio_record_schema")

BUILTIN_ALIASES = [
    "ma_validate",
    "ma",
    "ma_fields",
    "mu_fields",
    "mu_schemas",
]


@dataclasses.dataclass
class MarshmallowNode:
    schema: Dict
    parent: "MarshmallowNode"
    stack: ModelBuilderStack

    read: bool
    write: bool

    field_name: str
    exact_field: str
    field_class: str
    field_arguments: List[str]
    field_imports: List[Import]
    imports: List[Dict[str, str]]
    used: bool = False

    @classmethod
    def from_stack(
        cls, schema: ModelSchema, stack: ModelBuilderStack, marshmallow_field: str
    ):
        datatype = datatypes.get_datatype(
            stack.top.data, stack.top.key, schema.model, schema, stack
        )
        try:
            definition = getattr(datatype, marshmallow_field)()
        except Exception as e:
            raise InvalidModelException(
                f"Error getting marshmallow definition at path {stack.path}: top is {stack.top.data}"
            ) from e
        imports = datatype.imports()
        field_arguments = copy.copy(definition.get("arguments", []))
        validators = definition.get("validators", [])
        if validators:
            field_arguments.append(f"validate=[{', '.join(validators)}]")

        constructor_arguments = dict(
            schema=schema,
            parent=None,
            # semi-shallow clone
            stack=stack.clone(),
            field_arguments=field_arguments,
            field_imports=imports,
            **cls._kwargs(definition, schema, stack),
        )
        return cls(**constructor_arguments)

    @classmethod
    def _kwargs(cls, definition: Any, schema: ModelSchema, stack: ModelBuilderStack):
        field_name = definition.get("field-name", stack.top.key)
        field_name = convert_name_to_python(field_name)
        return {
            "read": definition.get("read", True),
            "write": definition.get("write", True),
            "field_name": field_name,
            "exact_field": definition.get("field", None),
            "field_class": definition.get("field-class", None),
            "imports": definition.get("imports", []),
        }

    def get_imports(self) -> List[Import]:
        if not self.exact_field:
            generated_field_imports: List[Import] = list(self.field_imports)
        else:
            generated_field_imports = []
        return [
            Import(k["import"], k.get("alias")) for k in self.imports
        ] + generated_field_imports

    def prepare(self, package_name: str):
        pass

    @property
    def field(self):
        return f"{self.field_name} = {self._field_rhs}"

    @property
    def _field_rhs(self):
        if self.exact_field:
            return self.exact_field
        else:
            rw_arguments = []
            if self.read and not self.write:
                rw_arguments.append("dump_only=True")
            elif self.write and not self.read:
                rw_arguments.append("load_only=True")
            if self.stack.top.key != self.field_name:
                rw_arguments.append(f'data_key="{self.stack.top.key}"')
                rw_arguments.append(f'attribute="{self.stack.top.key}"')
            return (
                f"{self.field_class}"
                + f"({', '.join(self.field_arguments + rw_arguments)})"
            )

    def mark_used(self):
        if self.read or self.write:
            self.used = True
        return self.used


@dataclasses.dataclass
class CompositeMarshmallowNode(MarshmallowNode):
    fields: List["MarshmallowNode"] = dataclasses.field(default_factory=list)

    @classmethod
    def _kwargs(cls, definition: Any, schema: ModelSchema, stack: ModelBuilderStack):
        return {"fields": [], **super()._kwargs(definition, schema, stack)}

    def add_field(self, field: "MarshmallowNode"):
        field.parent = self
        self.fields.append(field)

    def walk(self):
        for fld in self.fields:
            if isinstance(fld, CompositeMarshmallowNode):
                yield from fld.walk()
            else:
                yield fld
        yield self

    def mark_used(self):
        if self.read or self.write:
            used = False
            for fld in self.fields:
                if fld.mark_used():
                    used = True
            self.used = used
        return self.used


@dataclasses.dataclass
class ObjectMarshmallowNode(CompositeMarshmallowNode):
    generate: bool = True
    schema_class: Union[str, None] = None
    base_classes: Union[List[str], None] = None
    extra_fields: Union[List[Dict[str, str]], None] = None

    @classmethod
    def _kwargs(cls, definition: Any, schema: ModelSchema, stack: ModelBuilderStack):
        assert "schema-class" in definition, (
            f"Marshmallow schema class should be prepared by now, "
            + f"do you have datatypes model preprocessor configured? Definition: {stack.path}: {definition}: {stack.top.data}"
        )
        return {
            "generate": definition.get("generate", True),
            "schema_class": definition.get("schema-class", None),
            "base_classes": definition.get("base-classes", ["ma.Schema"]),
            "extra_fields": definition.get("extra-fields", []),
            **super()._kwargs(definition, schema, stack),
        }

    def prepare(self, package_name: str):
        super().prepare(package_name)
        if self.schema_class is None:
            return
        if self.exact_field:
            return
        self.field_arguments = [
            f"lambda: {split_base_name(self.schema_class)}()"
        ] + self.field_arguments
        self.field_imports = [
            *self.field_imports,
            Import(import_path=self.schema_class, alias=None),
        ]

    def generate_schema_class(self):
        if not self.used:
            return None, None, None
        if not self.generate:
            return None, None, None
        if not self.schema_class:
            return None, None, None

        field_list = []
        for fld in self.extra_fields:
            field_list.append(f"{fld.name} = {fld.value}")

        for fld in self.fields:
            if fld.read or fld.write:
                # and generate the field
                field_list.append(fld.field)

        package_name, base_class_name = split_package_base_name(self.schema_class)

        field_list = "\n    ".join(field_list)
        base_classes = []
        for x in self.base_classes:
            base_class_package_name, base_class_class_name = split_package_base_name(x)
            if base_class_package_name in BUILTIN_ALIASES:
                base_classes.append(x)
            else:
                base_classes.append(base_class_class_name)

        return (
            package_name,
            base_class_name,
            f"""
class {base_class_name}({", ".join(base_classes)}):
    \"""{base_class_name} schema.\"""
    {field_list}
""",
        )

    def get_imports(self) -> List[Import]:
        imports = super().get_imports()
        for base_class in self.base_classes:
            if "." in base_class:
                imports.append(Import(import_path=base_class, alias=None))
        return imports


@dataclasses.dataclass
class ArrayMarshmallowNode(CompositeMarshmallowNode):
    @property
    def _field_rhs(self):
        # wrap the child's RHS with list
        ret = self._first_child._field_rhs
        rw_arguments = []
        if self.stack.top.key != self.field_name:
            rw_arguments.append(f'data_key="{self.stack.top.key}"')
            rw_arguments.append(f'attribute="{self.stack.top.key}"')
        if rw_arguments:
            rw_arguments = ", " + ", ".join(rw_arguments)
        else:
            rw_arguments = ""
        return f"{self.field_class}({ret}{rw_arguments})"

    @property
    def field_imports(self) -> List[Import]:
        return self._first_child.field_imports

    @field_imports.setter
    def field_imports(self, imports):
        """Do not set imports as we generate the first child, not this node"""

    @property
    def _first_child(self):
        if len(self.fields) != 1:
            raise InvalidModelException(
                f"No array items or too many array items at {self.stack.path}"
            )
        return self.fields[0]


class InvenioRecordSchemaBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_schema"
    class_config = "record-schema-class"
    template = "record-schema"
    OAREPO_MARSHMALLOW_PROPERTY = "marshmallow"
    extra_imports = []

    marshmallow_stack: List[MarshmallowNode]

    def begin(self, schema, settings):
        super().begin(schema, settings)

        stack = ModelBuilderStack()
        stack.push("model", schema.current_model)

        self.marshmallow_stack = [
            ObjectMarshmallowNode.from_stack(
                schema.schema, stack, marshmallow_field=self.OAREPO_MARSHMALLOW_PROPERTY
            )
        ]

    def finish(self):
        super(PythonBuilder, self).finish()
        model_node = self.marshmallow_stack[0]

        # generate schema classes wherever they are not filled
        package_name = split_package_name(
            getattr(self.current_model, self.class_config)
        )

        generated_classes = defaultdict(list)
        generated_imports = defaultdict(set)

        model_node.mark_used()
        if model_node.generate:
            model_node.used = True

        known_classes = set()
        for fld in model_node.walk():
            fld.prepare(package_name)
            if hasattr(fld, "generate_schema_class"):
                package, class_name, generated_class = fld.generate_schema_class()
                if (package_name, class_name) not in known_classes:
                    known_classes.add((package_name, class_name))
                    if class_name:
                        imports = generated_imports[package]
                        for _f in fld.walk():
                            imports.update(_f.get_imports())
                        generated_classes[package].append(generated_class)

        # create python files ...
        for package_name in generated_classes:
            python_path = self.class_to_path(f"{package_name}.Dummy")
            x: Import
            imports = [
                x
                for x in generated_imports[package_name]
                if split_package_name(x.import_path)
                not in (
                    package_name,
                    # known prefixes
                    *BUILTIN_ALIASES,
                )
            ] + self.extra_imports

            imports = list(sorted(imports, key=lambda x: x.import_path))

            self.process_template(
                python_path,
                self.template,
                current_package_name=package_name,
                imports=imports,
                generated_classes=generated_classes[package_name],
            )
        # super.finish not called as it is handled in the for loop above

    @process("**", condition=lambda current, stack: stack.schema_valid)
    def enter_model_element(self):
        schema_element_type = self.stack.top.schema_element_type

        if schema_element_type in ("property", "items"):
            datatype = datatypes.get_datatype(
                self.stack.top.data,
                self.stack.top.key,
                self.schema.model,
                self.schema,
                self.stack,
            )
            json_schema_type = self.stack.top.json_schema_type
            if isinstance(datatype, ObjectDataType):
                node_class = ObjectMarshmallowNode
            elif json_schema_type == "array":
                node_class = ArrayMarshmallowNode
            else:
                # primitive value
                node_class = MarshmallowNode

            fld = node_class.from_stack(
                self.schema,
                self.stack,
                marshmallow_field=self.OAREPO_MARSHMALLOW_PROPERTY,
            )
            self.marshmallow_stack[-1].add_field(fld)
            self.marshmallow_stack.append(fld)
            self.build_children()
            self.marshmallow_stack.pop()
        else:
            self.build_children()



from collections import defaultdict


def assert_python_equals(actual, expected, msg=""):
    actual_lines = [x.strip() for x in actual.split("\n")]
    expected_lines = [x.strip() for x in expected.split("\n")]

    expected_by_content = defaultdict(list)
    for lineno, line in enumerate(expected_lines):
        line = line.strip()
        if not line:
            continue
        expected_by_content[line].append(lineno)

    last_line = 0
    for actual_line_no, actual_line in enumerate(actual_lines):
        actual_line = actual_line.strip()
        if not actual_line:
            continue
        if actual_line in expected_by_content and expected_by_content[actual_line]:
            last_line = expected_by_content[actual_line].pop(0)
        else:
            print("Expected lines:\n")
            print_lines_around(expected_lines, last_line)

            print("Actual lines:\n")
            print_lines_around(actual_lines, actual_line_no)

            raise AssertionError(
                f"Actual line {actual_line_no + 1} '{actual_line}' not in expected lines. {msg}"
            )


def print_lines_around(lines, position):
    for p in range(max(0, position - 10), min(len(lines), position + 11)):
        print(f"{str(p + 1):5s} {lines[p]}")



from oarepo_model_builder.datatypes import datatypes

from . import DeepTransformationModelPreprocessor


class DatatypeDefaultModelPreprocessor(DeepTransformationModelPreprocessor):
    def begin(self, schema, settings):
        super().begin(schema, settings)
        self.context = {}

    def transform_node(self, stack, data):
        if stack.top.schema_element_type in ("property", "items"):
            # stack top is property or items
            datatype = datatypes.get_datatype(
                data, stack.top.key, self.schema.current_model, self.schema, stack
            )
            if datatype:
                datatype.prepare(self.context)
